/*
    COEFX
    
    Routines for converting RUDAK telemetry channels to engineering units.
        
    Channel name and conversion formula info are read from a file in the
    following format:
    
    The file contains one or more subsystem blocks. Each subsystem block
    begins with a line containing the subsystem name. Each additional line
    in the block begins with a channel number in hex, a string naming the
    channel in 15 columns (including the trailing space), 3 or 6 floating
    point decimal numbers comprising the coefficients of the conversion
    polynomial, and a string naming the units for the channel. The special
    subsystem name "NOTES" marks the end of the data portion of the file,
    and the free text after that line is ignored.
    
    Bizarrely, and for no apparent reason, the coefficients are in different
    orders for the 3-term and 6-term formats. The 3-term format has the
    low-order (constant offset) term first, and the 6-term format has it
    last. Standards, learn to love them.

$Header: C:/Projects/RUDAKW~1/RCS/coefx.c,v 1.2 2001/05/24 01:00:22 ptw Exp ptw $

Copyright 2001, Paul Williamson.

2001/05/23 ptw Trim trailing spaces from units.

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "coefx.h"

#define SUBSYS_WIDTH    40          /* Maximum length of a subsystem ID */

#define NAME_WIDTH      15          /* Columns allocated to channel name */
                                    /* Includes a mandatory trailing space */
                                    
#define UNITS_WIDTH     15          /* Maximum width of the units field */

#define POLY_TERMS       6          /* Maximum number of terms in the polynomial */

typedef struct  {
    char    chan_name[NAME_WIDTH];  /* Name of the telemetry channel */
    double  poly[POLY_TERMS];       /* Coefficients of conversion polynomial */
    char    chan_units[UNITS_WIDTH];/* Units for this channel */
    int     exists;                 /* non-zero if this channel was read from
                                       the calibration file */
} chan_type;


static  chan_type   cal[256];
                                    /* Calibration values read from the file.
                                       There are at most 256 channels possible,
                                       since the channel number is only one
                                       byte.*/

/***************************************************************************
   read_calibration_file
   
   Given:
        the name of a calibration file, and
         
        an array of strings specifying which subsystems we're using,
        
        
   read the calibration for those subsystems into the cal array.
   May be called multiple times to register values from different
   coefficient files.
   
   It is a fatal error if the calibration file can't be read.
   
   A warning message is output to stderr if a channel is redefined,
   and the new definition overwrites the old one.
   
   Returns: the number of valid channels in the desired subsystems
            found in the file.
   
***************************************************************************/

#define MAXLINE 200
   
int read_calibration_file(char *calfilename, char **subsys_list)
{
    int     values_found = 0;
    char    linebuf[MAXLINE];
    char    token[SUBSYS_WIDTH];
    FILE    *calfile;
    int     channel;
    int     j;
    char    *subsys;

    chan_type   chan;                   /* Temp variable for the data we're
                                           trying to read in. */

    int     reading_data_lines = 0;     /* Flag: set after we've found a
                                           subsystem name we like, cleared
                                           when we find a line that isn't
                                           a valid data line OR a subsystem
                                           name we like. */

    if ((calfile = fopen(calfilename, "r")) == NULL) {
        fprintf(stderr, "Error opening calibration file %s: ", calfilename);
        perror("");
        exit(1);
    }

    while (1) {

        if (fgets(linebuf, MAXLINE, calfile) == NULL) {
            if (feof(calfile))
                return values_found;
            else {
                fprintf(stderr, "Error reading calibration file %s: ", calfilename);
                perror("");
                exit(1);
            }
        }

        if (strncmp(linebuf, "NOTES", 5) == 0)
            return values_found;

        if (reading_data_lines) {

            /* Try to match a 6-term data line */
            if (sscanf(linebuf, " %x %14c %le %le %le %le %le %le %14[^\n]",
                                &channel, &chan.chan_name,
                                &chan.poly[5], &chan.poly[4], &chan.poly[3],
                                &chan.poly[2], &chan.poly[1], &chan.poly[0],
                                &chan.chan_units) == 9) {
                if (cal[channel].exists)
                    fprintf(stderr, "Channel %02X redefined.\n", channel);
                chan.exists = 1;
                chan.chan_name[14] = '\0';
                for (j=13; j >= 0 && chan.chan_name[j] == ' '; j-- )
                    chan.chan_name[j] = '\0';
                for (j=strlen(chan.chan_units)-1; j >= 0; j--)
                    if (isspace(chan.chan_units[j]))
                        chan.chan_units[j] = '\0';
                    else
                        break;
                cal[channel] = chan;    /* Copy whole struct */
                continue;               /* Go on to the next line */
            }

            /* Try to match a 3-term data line */
            if (sscanf(linebuf, " %x %14c %le %le %le %14[^\n]", &channel,
                       &chan.chan_name, &chan.poly[0], &chan.poly[1],
                       &chan.poly[2], &chan.chan_units) == 6) {
                chan.poly[3] = 0.0;
                chan.poly[4] = 0.0;
                chan.poly[5] = 0.0;
                if (cal[channel].exists)
                    fprintf(stderr, "Channel %02X redefined.\n", channel);
                chan.exists = 1;
                chan.chan_name[14] = '\0';
                for (j=13; j >= 0 && chan.chan_name[j] == ' '; j-- )
                    chan.chan_name[j] = '\0';
                for (j=strlen(chan.chan_units)-1; j >= 0; j--)
                    if (isspace(chan.chan_units[j]))
                        chan.chan_units[j] = '\0';
                    else
                        break;
                cal[channel] = chan;    /* Copy whole struct */
                continue;               /* Go on to the next line */
            }
        }/* if (reading_data_lines) */
        
        /* If we fall through to here, we have found a line that isn't a
           valid data line. Note that this means we won't tolerate any
           blank lines within a subsystem's data lines. */
        reading_data_lines = 0;

        /* Try to match the desired subsystem names */
        j = 0;
        while ((subsys = subsys_list[j]) != NULL) {
            if (   sscanf(linebuf, "%40s", &token) == 1
                && strcmp(token, subsys) == 0 ) {

                /* We found the subsystem token for a desired subsystem */
                reading_data_lines = 1;
                break;  /* stop trying to match the subsystem name, and
                           go on to the next input line */
            }

            j++;        /* next candidate subsystem name */
        }
        
        /* If we fall through to here without setting reading_data_lines,
           we have found a line that's not a desired subsystem name. If the
           file format is valid, we're in an undesired subsystem. That's OK.
           If the file format is screwy, we might be skipping over garbage,
           but that's OK too.
           
           Just keep reading lines and trying to match a desired subsystem
           name until we find one (or until EOF). */

    }/* while(1) */

}/* read_calibration_file() */


/**************************************************************************
    supply_dummy_definition
    
    Given a channel number, fill its entry in the cal array with dummy
    values, as follows:
        chan_name = "UnknownXX" where XX is the channel number in hex
        chan_units = "counts"
        poly = the identity polynomial {0, 1, 0, 0, 0, 0}
**************************************************************************/

static void supply_dummy_definition(int channel) {

    int i;

    fprintf(stderr, "Uncalibrated channel %02X\n", channel);

    sprintf(cal[channel].chan_name, "Unknown%02X", channel);
    strcpy(cal[channel].chan_units, "counts");
    cal[channel].poly[0] = 0.0;
    cal[channel].poly[1] = 1.0;
    for (i=2; i < POLY_TERMS; i++)
        cal[channel].poly[i] = 0.0;
    cal[channel].exists = 1;

}/* supply_dummy_definition() */

/**************************************************************************
   counts_to_eng_units
   
   Given a telemetry value and its channel number, return its value in
   engineering units.
   
   The conversion formula is taken from the cal array.
      
   If no definition exists for the specified channel, an error message is
   output to stderr and a dummy definition is created for that channel.
**************************************************************************/   
double counts_to_eng_units(unsigned int counts, int channel) {

    double  value = 0.0;
    int     i;

    if (!cal[channel].exists)
        supply_dummy_definition(channel);

    for (i = POLY_TERMS-1; i > 0; i--) {        /* Horner method */
        value += cal[channel].poly[i];
        value *= counts;
    }
    value += cal[channel].poly[0];              /* Constant term */

    return value;

}/* counts_to_eng_units() */

/*************************************************************************
   channel_name
   
   Given a channel number, return a pointer to its name.
   
   The name is taken from the cal array.
      
   If no definition exists for the specified channel, a dummy definition
   is created for that channel. 
*************************************************************************/   
char *channel_name(int channel) {

    if (!cal[channel].exists)
        supply_dummy_definition(channel);

    return cal[channel].chan_name;

}/* channel_name() */


/************************************************************************
   eng_units
   
   Given a channel number, return a pointer to the string describing
   the engineering units used for that channel.  
   
   The name is taken from the cal array.
      
   If no definition exists for the specified channel, a dummy definition
   is created for that channel. 
*************************************************************************/   
char *eng_units(int channel) {

    if (!cal[channel].exists)
        supply_dummy_definition(channel);

    return cal[channel].chan_units;

}/* eng_units_units() */


/* Unit test framework */

#ifdef MAIN

int main(int argc, char **argv)
{
    int     i;

    char    *ss_list[] = { "RUDAKA", "RUDAKCANA", "RUDAKCANB" };



    read_calibration_file("RUDCOEFX", ss_list);

    for (i = 0; i < 256; i++) {
        if (cal[i].exists) {
            printf("%02X %14.14s %lf %lf %lf %lf %lf %lf %s\n",
                   i,
                   cal[i].chan_name,
                   cal[i].poly[0],
                   cal[i].poly[1],
                   cal[i].poly[2],
                   cal[i].poly[3],
                   cal[i].poly[4],
                   cal[i].poly[5],
                   cal[i].chan_units);
        }
    }

    return 0;

}

#endif /* MAIN */

